""" Author: Philipp Steinrötter (steinroe) """

from requests.adapters import HTTPAdapter

import json
import requests
import ssl

from .response import Response
from .utils import current_milli_time


class RESTGatewayException(Exception):
    pass


class RestClient(object):

    def __init__(self, instance: str, device_alternate_id: str, pemfile: str, secret: str):
        """Instantiate REST Client configured for specified instance and device
        
        Arguments:
            instance {str} -- IoT Services instance
            device_alternate_id {str} -- The alternate id of the device
            certfile_path {str} -- The certfile path for the device
            keyfile_path {str} -- The keyfile path for the device
        """

        self.instance = instance
        self.device_alternate_id = device_alternate_id
        self.pemfile = pemfile

        self.session = self._init_session(pemfile, secret)

        self.gateway_uri = '/iot/gateway'

    def _init_session(self, pemfile: str, secret: str):
        session = requests.Session()
        adapter = RESTGatewayAdapter(pemfile=pemfile, secret=secret)
        session.mount('https://' + self.instance, adapter)
        return session

    def _request_gateway(self, service: str, headers: dict, payload: str) -> Response:
        """Fires a HTTP request against a gateway
        
        Arguments:
            service {str} -- Service path
            headers {dict} -- HTTP headers
            payload {str} -- Message payload
        
        Returns:
            Response -- Response object
        """

        service = 'https://' + self.instance + self.gateway_uri + '/rest' + service

        try:
            response = self.session.request(method='POST', url=service, data=payload, headers=headers)
            response.raise_for_status()
            if response.status_code == 207:
                # Batch upload with partial failure. Raise Exception.
                raise RESTGatewayException(self._parse_error(json.loads(response.text)))
            return Response(response.status_code, json.loads(response.text), response.headers)
        except requests.exceptions.HTTPError as err:
            raise RESTGatewayException(self._parse_error(json.loads(err.response.text)))

    def _parse_error(self, message_infos) -> str:
        messages = ''
        for msg in message_infos:
            if msg.get('messages') is not None:
                messages += ' '.join(msg['messages'])

    def post_command(self, capability_alternate_id: str, sensor_alternate_id: str, command: dict, device_alternate_id: str = None) -> Response:
        """Post commands over rest gateway for specified device
        
        Arguments:
            device_alternate_id {str} -- Alternate ID of the device. If none, the device from the client id will be used.
            capability_alternate_id {str} -- Alternate ID for capability
            sensor_alternate_id {str} -- Alternate ID for sensor
            command {dict} -- Dict with the keys and respective values of the desired commands
        
        Returns:
            Response -- Response object
        """
        if device_alternate_id is None:
            device_alternate_id = self.device_alternate_id

        service = '/commands/' + device_alternate_id
        headers = {'Content-Type': 'application/json'}
        payload_json = json.dumps(
            {"capabilityAlternateId": capability_alternate_id, "sensorAlternateId": sensor_alternate_id,
             "command": command})
        response = self._request_gateway(service=service, headers=headers, payload=payload_json)
        return response

    def post_measures(self, capability_alternate_id: str, sensor_alternate_id: str, measures: list,
                      sensor_type_alternate_id: int = None, timestamp: int = None, device_alternate_id: str = None) -> Response:
        """Post measures over rest gateway for specified device
        
        Arguments:
            capability_alternate_id {str} -- Alternate ID for capability
            sensor_alternate_id {str} -- Alternate ID for sensor
            measures {list} -- List of key-value pairs with the keys and respective values of the desired measures
            sensor_type_alternate_id {int} -- (Optional) If this parameter is set, the device will be auto-onboarded if it does not exist yet. Note: The alternate id of the sensor type must be numeric.
            timestamp {int} -- UNIX time in milliseconds. If None, current time will be used.
            device_alternate_id {str} -- Alternate ID of the device. If none, the device from the client id will be used.
        Returns:
            Response -- Response object
        """
        if device_alternate_id is None:
            device_alternate_id = self.device_alternate_id

        service = '/measures/' + device_alternate_id
        headers = {'Content-Type': 'application/json'}
        payload = {
            "timestamp": current_milli_time(),
            "capabilityAlternateId": capability_alternate_id,
            "sensorAlternateId": sensor_alternate_id,
            "measures": measures
        }

        if sensor_type_alternate_id is not None:
            payload['sensorTypeAlternateId'] = sensor_type_alternate_id

        if timestamp is not None:
            payload['timestamp'] = timestamp

        payload_json = json.dumps(payload)
        response = self._request_gateway(service=service, headers=headers, payload=payload_json)
        return response

    def post_batched_measures(self, messages: list, device_alternate_id: str = None) -> Response:
        """Post batched measures over rest gateway
        
        Arguments:
            messages {list} -- List of dicts, each containing sensorAlternateId, capabilityAlternateId and an array of measures as key-value pairs. If the device should be onboarded automatically, the sensorTypeAlternateId must also be provided.
            device_alternate_id {str} -- Alternate ID of the device. If none, the device from the client id will be used.

        Returns:
            Response -- Response object
        """
        if device_alternate_id is None:
            device_alternate_id = self.device_alternate_id

        service = '/measures/' + device_alternate_id
        headers = {'Content-Type': 'application/json'}
        payload_json = json.dumps(messages)
        response = self._request_gateway(service=service, headers=headers, payload=payload_json)
        return response


class RESTGatewayAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        pemfile = kwargs.pop('pemfile', None)
        secret = kwargs.pop('secret', None)
        if pemfile is None or secret is None:
            raise ValueError('Either "pemfile" or "secret" is missing')
        self.ssl_context = self._create_ssl_context(pemfile, secret)
        super(RESTGatewayAdapter, self).__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        if self.ssl_context:
            kwargs['ssl_context'] = self.ssl_context
        return super(RESTGatewayAdapter, self).init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        if self.ssl_context:
            kwargs['ssl_context'] = self.ssl_context
        return super(RESTGatewayAdapter, self).proxy_manager_for(*args, **kwargs)

    def _create_ssl_context(self, pemfile, secret):
        if ssl is None:
            raise ValueError('This platform has no SSL/TLS.')
        if not hasattr(ssl, 'SSLContext'):
            # Require Python version that has SSL context support in standard library
            raise ValueError('Python 2.7.9 and 3.2 are the minimum supported versions for TLS.')
        # Create SSLContext object
        tls_version = ssl.PROTOCOL_TLSv1
        # If the python version supports it, use highest TLS version automatically
        if hasattr(ssl, "PROTOCOL_TLS"):
            tls_version = ssl.PROTOCOL_TLS
        context = ssl.SSLContext(tls_version)

        # Configure context
        if pemfile is not None:
            context.load_cert_chain(pemfile, password=secret)

        return context



""" Author: Philipp Steinrötter (steinroe) """

import json

from .tenant_iot_service import TenantIoTService
from .mqtt_client import MQTTClient
from .rest_client import RestClient
from .utils import build_query
from .response import Response


class DeviceService(TenantIoTService):
    def __init__(self,
                 instance,
                 user,
                 password,
                 tenant_id):
        """Instantiate DeviceService object
        
        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
            tenant_id {string} -- Id of the tenant
        """

        self.service = '/devices'

        TenantIoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password,
            tenant_id=tenant_id
        )

    def get_devices(self, filters=None, orderby=None, asc=True, skip=None, top=None) -> Response:
        """The endpoint returns a list of devices.
        
        Keyword Arguments:
            filters {list} -- This parameter allows clients to filter the collection for specific attributes. It is possible to filter by 'id’, 'alternateId’, 'gatewayId’, 'name’, 'description’, and 'status’. The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"]. (default: {None})
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore included in the result set (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request (default: {None})
        
        Returns:
            Response -- Response object
        """
        query = build_query(filters=filters, orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=self.service, headers=None, payload=None, accept_json=True,
                                    query=query)

    def create_device(self, gateway_id: str, name: str, as_router=False, custom_properties=None) -> Response:
        """This endpoint is used to create a device.
        
        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            name {str} -- Unique identifier of a name
            custom_properties {list} -- List of dicts with the keys 'key' and 'value' specifying the custom properties

        Returns:
            Response --  Response object
        """
        headers = {'Content-Type': 'application/json'}
        payload = {"gatewayId": gateway_id, "name": name, "customProperties": custom_properties}
        if as_router is True:
            payload['authorizations'] = [{'type': 'router'}]
        payload_json = json.dumps(payload)
        return super().request_core(method='POST', service=self.service, headers=headers, payload=payload_json,
                                    accept_json=True)

    def get_device_count(self) -> Response:
        """The endpoint returns the count of all devices.

        Returns:
            Response -- Response object
        """
        service = self.service + '/count'
        return super().request_core(method='GET', service=service, accept_json=True)

    def delete_device(self, device_id: str) -> Response:
        """The endpoint is used to delete the device associated to the given id.

        Arguments:
            device_id {str} -- Unique identifier of a device
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id
        return super().request_core(method='DELETE', service=service)

    def get_device(self, device_id: str) -> Response:
        """The endpoint returns the device associated to the given id.
        
        Arguments:
            device_id {str} -- Unique identifier of a device
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id
        return super().request_core(method='GET', service=service, accept_json=True)

    def update_device(self, device_id: str, name: str) -> Response:
        """This endpoint is used to update the device associated to the given id with details specified in the request body. This endpoint can only be used to modify a devices name. To update custom properties, sensors or authentications, use the respective APIs.
        
        Arguments:
            device_id {str} -- Unique identifier of a device
            name {str} -- New device name
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id
        headers = {'Content-Type': 'application/json'}
        payload = '{ "name" : "' + name + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def get_device_certs(self, device_id: str) -> Response:
        """The endpoint is used to list the fingerprints and expiration dates for device certificates of the given device.

        Arguments:
            device_id {str} -- Unique identifier of a device

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id + '/authentications/clientCertificate'
        return super().request_core(method='GET', service=service, accept_json=True)

    def get_device_p12(self, device_id: str) -> Response:
        """The endpoint is used to download device specific p12 file for authentication.
        
        Arguments:
            device_id {str} -- Unique identifier of a device
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id + '/authentications/clientCertificate/p12'
        return super().request_core(method='GET', service=service, accept_json=True)

    def get_device_pem(self, device_id: str) -> Response:
        """Used to download a device specific private key and certificate in PEM format for authentication.
        
        Arguments:
            device_id {str} -- Unique identifier of a device
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id + '/authentications/clientCertificate/pem'
        return super().request_core(method='GET', service=service, accept_json=True)

    def create_device_pem(self, device_id: str, csr: str) -> Response:
        """The endpoint is used to create a device specific certificate in PEM format for authentication.

        Arguments:
            device_id {str} -- Unique identifier of a device
            request {dict} -- Specification of the certificate signing request for the device certificate.

        Returns:
            Response -- Response object
        """
        # service = self.service + '/' + device_id + '/authentications/clientCertificate/pem'
        # headers = {'Content-Type': 'application/json'}
        # payload = json.dumps({
        #     'csr': csr,
        #     'type': 'clientCertificate'
        # })
        # return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)
        raise NotImplementedError('This method is not supported yet.')

    def revoke_device_cert(self, device_id: str, fingerprint: str) -> Response:
        """The endpoint is used to revoke a certificate of the given device.

        Arguments:
            device_id {str} -- Unique identifier of a device
            fingerprint {str} -- The fingerprint of the certificate hashed with SHA-256 in hex format.

        Returns:
            Response -- Response object
        """
        # service = self.service + '/' + device_id + '/authentications/clientCertificate/' + fingerprint
        # return super().request_core(method='DELETE', service=service)
        raise NotImplementedError('This method is not supported yet.')

    def send_command_to_device(self, device_id: str, capability_id: str, sensor_id: str, command: dict) -> Response:
        """Used to send the command specified in the request body to the device associated to the given id.
        
        Arguments:
            device_id {str} -- Unique identifier of a device
            capability_id {str} -- Unique identifier of a capability
            sensor_id {str} -- Unique identifier of a sensor
            command {dict} --  Dict with additional properties and their respective values
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id + '/commands'
        headers = {'Content-Type': 'application/json'}
        payload = json.dumps({"capabilityId": capability_id, "sensorId": sensor_id, "command": command})
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def add_custom_property(self, device_id: str, key: str, value: str) -> Response:
        """Used to add a custom property to the device associated to the given id.
        
        Arguments:
            device_id {str} --  Unique identifier of a device
            key {str} -- Key of the custom property
            value {str} -- Value of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id + '/customProperties'
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def delete_custom_property(self, device_id: str, key: str) -> Response:
        """Delete a custom property from the device associated to the given id.
        
        Arguments:
            device_id {str} -- Unique identifier of a device
            key {str} -- Key of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id + '/customProperties/' + key
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def update_custom_property(self, device_id: str, key: str, value: str) -> Response:
        """Updates a custom property of the device associated to the given id. The ‘key’ attribute cannot be modified.
        
        Arguments:
            device_id {str} -- Unique identifier of a device
            key {str} -- Key of the custom property
            value {str} -- The updated value of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id + '/customProperties/' + key
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def get_mqtt_client(self, device_alternate_id: str, pemfile: str, secret: str) -> MQTTClient:
        """Returns MQTT client for specified router device
        
        Arguments:
            device_alternate_id {str} -- Alternate identifier of a device
            certfile_path {str} -- Path to the certfile
            keyfile_path {str} -- Path to the keyfile
        
        Returns:
            MQTTClient -- The MQTTClient object configured for the specified router device
        """
        return MQTTClient(self.instance, device_alternate_id, pemfile, secret)

    def get_rest_client(self, device_alternate_id: str, pemfile: str, secret: str) -> RestClient:
        """Returns REST client for specified device
        
        Arguments:
            device_alternate_id {str} -- Alternate identifier of a device
            certfile_path {str} -- Path to the certfile
            keyfile_path {str} -- Path to the keyfile
        
        Returns:
            RestClient -- The RestClient object configured for the specified device
        """
        return RestClient(self.instance, device_alternate_id, pemfile, secret)

    def get_measures(self, device_id: str, filters=None, orderby=None, asc=True, skip=None, top=None) -> Response:
        """Returns a list of measures related to the device associated to the given id.
        
        Arguments:
            device_id {str} -- Unique identifier of a device
        
        Keyword Arguments:
            filter {list} -- This parameter allows clients to filter the collection for specific attributes. It is possible to filter by ‘capabilityId’ and 'timestamp’. When filtering by ‘timestamp’ the following binary operator are supported 'le’, 'lt’, 'ge’, and 'gt’. The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"] (default: {None})
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore included in the result set (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request (default: {None})
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + device_id + '/measures'
        query = build_query(filters=filters, orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=service, query=query, accept_json=True)



""" Author: Philipp Steinrötter (steinroe) """

import requests
import json
from .response import Response


class DeviceManagementAPIException(Exception):
    pass


class IoTService(object):
    def __init__(self, instance: str, user: str, password: str):
        """Instantiate IoT Service object
        
        Arguments:
            instance {str} -- IoT Service instance
            user {str} -- IoT Service user
            password {str} -- IoT Service password
        
        Raises:
            ValueError -- Raised if any argument is not provided
        """
        if instance is None or user is None or password is None:
            raise ValueError('You must specify your instance, user and password.')

        self.instance = instance
        self.user = user
        self.password = password

        self._api_path = '/iot/core/api/v1'

    def request_core(self, method=None, service=None, headers=None, payload=None, accept_json=False, query=None,
                     files=None) -> Response:
        """Fires a HTTP request to core services
        
        Keyword Arguments:
            method {str} -- HTTP method (default: {None})
            service {str} -- Service Path (default: {None})
            headers {dict} -- HTTP headers (default: {None})
            payload {str} -- Message payload (default: {None})
            accept_json {bool} -- If set to true, the response is parsed a JSON (default: {False})
            query {str} -- Query for filtering (default: {None})
            files {str} -- Path to files (default: {None})
        
        Returns:
            Response -- Response object
        """

        url = 'https://' + self.instance + self._api_path + service
        if query is not None:
            url = url + query

        user = self.user
        password = self.password

        try:
            response = requests.request(method, url, headers=headers, auth=(user, password), data=payload, files=files)
            response.raise_for_status()
            if accept_json:
                response_json = json.loads(response.text)
                return Response(response.status_code, response_json, response.headers)
            else:
                return Response(response.status_code, response.text, response.headers)
        except requests.exceptions.HTTPError as err:
            try:
                raise DeviceManagementAPIException(json.loads(err.response.text)['message'])
            except json.decoder.JSONDecodeError:
                raise DeviceManagementAPIException(err)



""" Author: Philipp Steinrötter (steinroe) """


class Response(object):
    """Objects contain information received from API"""

    def __init__(self, status_code, response, headers):
        self._status_code = status_code
        self._result = response
        self._headers = headers

    def get_result(self) -> str:
        """Returns the result of the response, e.g. the body of the message

        Returns:
            str -- The body of the response message. Mostly in JSON formatting.
        """
        return self._result

    def get_headers(self) -> str:
        """Returns the header of the response

        Returns:
            str -- The header of the response message.
        """
        return self._headers

    def get_status_code(self) -> int:
        """Status code of the response

        Returns:
            str -- The status code of the HTTP communication
        """
        return self._status_code



""" Author: Philipp Steinrötter (steinroe) """

import time
import logging
from http.client import HTTPConnection


def build_query(filters=None, orderby=None, asc=True, skip=None, top=None) -> str:
    """Builds query string
    
    Keyword Arguments:
        filters {list} -- The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"]. (default: {None})
        orderby {str} -- The attribute to order by. (default: {None})
        asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered ascending or descending. (default: {True})
        skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore included in the result set. (default: {None})
        top {int} -- This parameter restricts the maximum number of items which will be returned by the request. [description] (default: {None})
    
    Returns:
        str -- Query string
    """
    query_string = '?'
    if filters is not None and len(filters) > 0:
        query_string += 'filter='
        for idx, filter_query in enumerate(filters):
            query_string += filter_query
            if idx < len(filters) - 1:
                query_string += ' and '
    if orderby is not None:
        if len(query_string) is not 1:
            query_string += '&'
        query_string += 'orderby=' + orderby
        if asc is True:
            query_string += ' asc'
        else:
            query_string += ' desc'
    if skip is not None:
        if len(query_string) is not 1:
            query_string += '&'
        query_string += 'skip=' + skip
    if top is not None:
        if len(query_string) is not 1:
            query_string += '&'
        query_string += 'top=' + top

    if query_string != '?':
        return query_string
    else:
        return None


def current_milli_time():
    return int(round(time.time() * 1000))


def debug_requests_on():
    """Switches on logging of the requests module.
    """
    HTTPConnection.debuglevel = 1

    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


def debug_requests_off():
    """Switches off logging of the requests module.
    """
    HTTPConnection.debuglevel = 0

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.WARNING)
    root_logger.handlers = []
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.WARNING)
    requests_log.propagate = False



""" Author: Philipp Steinrötter (steinroe) """
import json

from .tenant_iot_service import TenantIoTService
from .utils import build_query
from .response import Response


class GatewayService(TenantIoTService):
    def __init__(self,
                 instance,
                 user,
                 password,
                 tenant_id):
        """Instantiate GatewayService object
        
        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
            tenant_id {string} -- Id of the tenant
        """

        self.service = '/gateways'

        TenantIoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password,
            tenant_id=tenant_id
        )

    def get_gateways(self, filters=None, orderby=None, asc=True, skip=None, top=None) -> Response:
        """The endpoint returns a list of gateways.
        
        Keyword Arguments:
            filters {list} -- This parameter allows clients to filter the collection for specific attributes. The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"] (default: {None})
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore not included in the result set (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request (default: {None})
        
        Returns:
            Response -- Response object
        """
        query = build_query(filters=filters, orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=self.service, query=query, accept_json=True)

    def get_gateway_count(self):
        """The endpoint returns the count of all gateways.

        Returns:
            Response -- Response object
        """
        service = self.service + '/count'
        return super().request_core(method='GET', service=service, accept_json=True)

    def delete_gateway(self, gateway_id: str) -> Response:
        """The endpoint is used to delete the gateway associated to the given id.
        
        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def get_gateway(self, gateway_id: str) -> Response:
        """The endpoint returns the gateway associated to the given id.
        
        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id
        return super().request_core(method='GET', service=service, accept_json=True)

    def update_gateway_name(self, gateway_id: str, name: str) -> Response:
        """The endpoint is used to update the gateway associated to the given id with details specified in the request body. To update custom properties, bundles or configuration, use the respective APIs.
        
        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            name {str} -- Name for the gateway
        
        Returns:
            Response -- Response object
        """

        service = self.service + '/' + gateway_id
        headers = {'Content-Type': 'application/json'}
        payload = '{ "name" : "' + name + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def get_gateway_certs(self, gateway_id: str) -> Response:
        """The endpoint is used to list the fingerprints and expiration dates for device certificates of the given gateway.

        Arguments:
            gateway_id {str} -- Unique identifier of a gateway

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/authentications/clientCertificate'
        return super().request_core(method='GET', service=service, accept_json=True)

    def create_gateway_pem(self, gateway_id: str, request: dict) -> Response:
        """The endpoint is used to create a gateway specific certificate in PEM format for authentication.

        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            request {dict} -- Specification of the certificate signing request for the device certificate.

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/authentications/clientCertificate/pem'
        headers = {'Content-Type': 'application/json'}
        payload = json.dumps(request)
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def revoke_gateway_cert(self, gateway_id: str, fingerprint: str) -> Response:
        """The endpoint is used to revoke a certificate of the given gateway.

        Arguments:
            gateway_id {str} -- Unique identifier of a device
            fingerprint {str} -- The fingerprint of the certificate hashed with SHA-256 in hex format.

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/authentications/clientCertificate/' + fingerprint
        return super().request_core(method='DELETE', service=service)

    def get_gateway_osgi_bundles(self, gateway_id: str, filters=None, orderby=None, asc=True, skip=None,
                                 top=None) -> Response:
        """The endpoint is used to revoke a certificate of the given gateway.

        Arguments:
            gateway_id {str} -- Unique identifier of a device
            filters {list} -- This parameter allows clients to filter the collection for specific attributes. The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"] (default: {None})
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore not included in the result set (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request (default: {None})

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/bundles'
        query = build_query(filters=filters, orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=service, query=query, accept_json=True)

    def install_gateway_osgi_bundle(self, gateway_id: str, osgi_bundle: str) -> Response:
        """This endpoint is used to install an OSGi bundle on the gateway associated to the given id. Note that bundles with a file size over 128 MB will be rejected. The installation takes place asynchronously: the provided bundle is stored on the system (where it is kept up to 24 hours), then a request to download it is dispatched to Gateway. As soon as it receives the request, Gateway initiates the bundle download. The API returns immediately after the download request is dispatched to Gateway; in order to inspect the outcome of the bundle installation, get_gateway_osgi_bundles() should be used

        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            osgi_bundle {str} -- Gateway OSGi Bundle

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/bundles'
        response = self.request_core(method='POST', service=service, accept_json=True, files=osgi_bundle)
        return response

    def delete_osgi_bundle(self, gateway_id: str, bundle_id: str) -> Response:
        """This endpoint is used to remove an OSGi bundle from the gateway associated to the given id.

        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            bundle_id {str} -- Unique identifier of an OSGi bundle

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/bundles/' + bundle_id
        response = self.request_core(method='DELETE', service=service, accept_json=True)
        return response

    def get_osgi_bundle(self, gateway_id: str, bundle_id: str) -> Response:
        """The endpoint returns the OSGi bundle associated to the given id.

        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            bundle_id {str} -- Unique identifier of an OSGi bundle

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/bundles/' + bundle_id
        response = self.request_core(method='GET', service=service, accept_json=True)
        return response

    def start_gateway_osgi_bundle(self, gateway_id: str, bundle_id: str) -> Response:
        """This endpoint is used to start the OSGi bundle of the gateway associated to the given id.

        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            bundle_id {str} -- Unique identifier of an OSGi bundle

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/bundles/' + bundle_id + '/start'
        response = self.request_core(method='POST', service=service, accept_json=True)
        return response

    def stop_gateway_osgi_bundle(self, gateway_id: str, bundle_id: str) -> Response:
        """This endpoint is used to stop the OSGi bundle of the gateway associated to the given id.

        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            bundle_id {str} -- Unique identifier of an OSGi bundle

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/bundles/' + bundle_id + '/stop'
        response = self.request_core(method='POST', service=service, accept_json=True)
        return response

    def get_gateway_configuration(self, gateway_id: str) -> Response:
        """The endpoint is used to download the gateway specific configuration XML file.
        
        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/configuration'
        return super().request_core(method='GET', service=service)

    def update_gateway_configuration(self, gateway_id: str, xml: str) -> Response:
        """The endpoint is used to update the gateway specific configuration by uploading a configuration XML file.
        
        Arguments:
            gateway_id {str} -- The endpoint is used to update the gateway specific configuration by uploading a configuration XML file.
            xml {str} -- XML file as string containing the Gateway configuration
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/configuration'
        headers = {'Content-Type': 'application/xml'}
        payload = xml
        return super().request_core(method='PUT', service=service, payload=payload, headers=headers)

    def add_custom_property(self, gateway_id: str, key: str, value: str) -> Response:
        """The endpoint is used to add a custom property to the gateway associated to the given id.
        
        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            key {str} -- Key of the custom property
            value {str} -- Value of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/customProperties'
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='POST', service=service, payload=payload, headers=headers, accept_json=True)

    def delete_custom_property(self, gateway_id: str, key: str) -> Response:
        """This endpoint is used to delete a custom property from the gateway associated to the given id.
        
        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            key {str} -- Key of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/customProperties/' + key
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def update_custom_property(self, gateway_id: str, key: str, value: str) -> Response:
        """The endpoint is used to update a custom property of the gateway associated to the given id. The ‘key’ attribute cannot be modified.

        Arguments:
            gateway_id {str} -- Unique identifier of a gateway
            key {str} -- Key of the custom property
            value {str} -- Updates value of the custom property
        
        Returns:
            Response -- Response object
        """

        service = self.service + '/' + gateway_id + '/customProperties/' + key
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def get_gateway_device_certs(self, gateway_id: str) -> Response:
        """The endpoint is used to list the fingerprints and expiration dates for device registration certificates of the given gateway.

        Arguments:
            gateway_id {str} -- Unique identifier of a device

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/authentications/clientCertificate'
        return super().request_core(method='GET', service=service, accept_json=True)

    def get_gateway_device_p12(self, gateway_id: str) -> Response:
        """The endpoint is used to download gateway specific p12 file for the registration of a device.

        Arguments:
            gateway_id {str} -- Unique identifier of a device

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/authentications/clientCertificate/p12'
        return super().request_core(method='GET', service=service, accept_json=True)

    def get_gateway_device_pem(self, gateway_id: str) -> Response:
        """The endpoint is used to download gateway specific pem file for the registration of a device.

        Arguments:
            gateway_id {str} -- Unique identifier of a device

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/authentications/clientCertificate/pem'
        return super().request_core(method='GET', service=service, accept_json=True)

    def revoke_gateway_device_cert(self, gateway_id: str, fingerprint: str) -> Response:
        """The endpoint is used to revoke a device registration certificate of the given gateway.

        Arguments:
            gateway_id {str} -- Unique identifier of a device
            fingerprint {str} -- The fingerprint of the certificate hashed with SHA-256 in hex format.

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + gateway_id + '/authentications/clientCertificate/' + fingerprint
        return super().request_core(method='DELETE', service=service)



""" Author: Philipp Steinrötter (steinroe) """

from .iot_service import IoTService
from .response import Response


class AboutService(IoTService):
    def __init__(self,
                 instance,
                 user,
                 password):
        """Instantiate AboutService object

        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
        """

        self.service = '/about'

        IoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password
        )

    def get_information(self) -> Response:
        """The endpoint returns information about the service and its configuration parameters.
        
        Returns:
            Response -- Response object
        """
        response = self.request_core(method='GET', service=self.service, accept_json=True)
        return response



""" Author: Philipp Steinrötter (steinroe) """

import json

from .tenant_iot_service import TenantIoTService
from .utils import build_query
from .response import Response


class CapabilityService(TenantIoTService):
    def __init__(self,
                 instance,
                 user,
                 password,
                 tenant_id):
        """Instantiate CapabilityService object
        
        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
            tenant_id {string} -- Id of the tenant
        """

        self.service = '/capabilities'

        TenantIoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password,
            tenant_id=tenant_id
        )

    def get_capabilities(self, filters=None, orderby=None, asc=True, skip=None, top=None) -> Response:
        """The endpoint returns a list of capabilities.
        
        Keyword Arguments:
            filters {list} -- This parameter allows clients to filter the collection for attributes of a capability. The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"] (default: {None})
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore not included in the result set (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request (default: {None})
        
        Returns:
            Response -- Response object
        """

        query = build_query(filters=filters, orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=self.service, headers=None, payload=None, accept_json=True,
                                    query=query)

    def create_capability(self, alternate_id: str, name: str, properties: list) -> Response:
        """This endpoint is used to create a capability.
        
        Arguments:
            alternate_id {str} -- Alternate ID of the capability
            name {str} -- Name of the capability
            properties {list} -- List of dicts describing the properties
        
        Returns:
            Response -- Response object
        """
        headers = {'Content-Type': 'application/json'}
        payload = json.dumps({"alternateId": alternate_id, "name": name, "properties": properties})
        return super().request_core(method='POST', service=self.service, headers=headers, payload=payload,
                                    accept_json=True)

    def get_capability_count(self) -> Response:
        """The endpoint returns the count of all capabilities.

        Returns:
            Response -- Response object
        """
        service = self.service + '/count'
        return super().request_core(method='GET', service=service, accept_json=True)

    def delete_capability(self, capability_id: str) -> Response:
        """The endpoint is used to delete the capability associated to the given id.
        
        Arguments:
            capability_id {str} -- Unique identifier of a capability
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + capability_id
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def get_capability(self, capability_id: str) -> Response:
        """The endpoint returns the capability associated to the given id.
        
        Arguments:
            capability_id {str} -- Unique identifier of a capability.
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + capability_id
        return super().request_core(method='GET', service=service, accept_json=True)

    def update_capability(self, capability_id: str, alternate_id: str = None, name: str = None, properties: list = None) -> Response:
        """This endpoint is used to update the capability associated to the given id with details specified in the request body.
        
        Arguments:
            capability_id {str} -- Unique identifier of a capability
            alternate_id {str} -- Alternate identifier of a capability
            name {str} -- Name of a capability
            properties {list} -- Properties of the capability
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + capability_id
        headers = {'Content-Type': 'application/json'}
        payload = {}
        if alternate_id is not None:
            payload['alternateId'] = alternate_id
        if name is not None:
            payload['name'] = name
        if properties is not None:
            payload['properties'] = properties
        payload_json = json.dumps(payload)
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload_json, accept_json=True)



""" Author: Philipp Steinrötter (steinroe) """

from .iot_service import IoTService
from .utils import build_query
from .response import Response


class ProtocolService(IoTService):
    def __init__(self,
                 instance,
                 user,
                 password):
        """Instantiate ProtocolService object
        
        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
        """
        self.service = '/protocols'

        IoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password
        )

    def get_protocols(self, skip: int, top: int) -> Response:
        """The endpoint returns a list of protocols available on the instance.
        
        Arguments:
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore included in the result
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request
        
        Returns:
            Response -- Response object
        """
        query = build_query(skip=skip, top=top)
        return super().request_core(method='GET', service=self.service, query=query, accept_json=True)

    def get_protocol_count(self):
        """The endpoint returns the count of all protocols.

        Returns:
            Response -- Response object
        """
        service = self.service + '/count'
        return super().request_core(method='GET', service=service, accept_json=True)

    def create_protocol(self, protocol_id: str) -> Response:
        """he endpoint is used to create a protocol. The new protocol is visible for all tenants of the instance.
        
        Arguments:
            protocol_id {str} -- ID of the protocol that will be created
        
        Returns:
            Response -- Response object
        """
        headers = {'Content-Type': 'application/json'}
        payload = '{ "id" : "' + protocol_id + '" }'
        return super().request_core(method='POST', service=self.service, headers=headers, payload=payload,
                                    accept_json=True)

    def delete_protocol(self, protocol_id: str) -> Response:
        """The endpoint is used to delete a protocol. Deleting is possible only if no tenant on the instance is using it anymore.
        
        Arguments:
            protocol_id {str} -- Unique identifier of a protocol.
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + protocol_id
        return super().request_core(method='DELETE', service=service, accept_json=True)



""" Author: Philipp Steinrötter (steinroe) """

import json

from .iot_service import IoTService
from .utils import build_query
from .response import Response


class TenantService(IoTService):
    def __init__(self,
                 instance,
                 user,
                 password):
        """Instantiate TenantService object
        
        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
        """

        self.service = '/tenants'

        IoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password
        )

    def get_tenants(self, filters=None, orderby=None, asc=True, skip=None, top=None) -> Response:
        """The endpoint returns a list of tenants.
        
        Keyword Arguments:
            filters {list} -- This parameter allows clients to filter the collection for attributes of a tenant. The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"]. (default: {None})
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore not included in the result set. (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request. (default: {None})
        
        Returns:
            Response -- Response object
        """
        query = build_query(filters=filters, orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=self.service, query=query, accept_json=True)

    def create_tenant(self, name: str, custom_properties=[]) -> Response:
        """The endpoint is used to create a tenant.
        
        Arguments:
            name {str} -- Name of the tenant
            custom_properties {list} -- Custom properties of the tenant as a list of dicts, each with the key-value pairs 'key' and 'value'
        
        Returns:
            Response -- Response object
        """
        headers = {'Content-Type': 'application/json'}
        payload = json.dumps({"name": name, "customProperties": custom_properties})
        return super().request_core(method='POST', service=self.service, headers=headers, payload=payload,
                                    accept_json=True)

    def get_tenants_count(self) -> Response:
        """Returns the count of all tenants.
        Returns:
            Response -- Response object
        """
        service = self.service + '/count'
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def delete_tenant(self, tenant_id: str) -> Response:
        """The endpoint is used to delete the tenant associated to the given id.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def get_tenant(self, tenant_id: str) -> Response:
        """The endpoint returns the tenant associated to the given id.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id
        return super().request_core(method='GET', service=service, accept_json=True)

    def update_tenant(self, tenant_id: str, name: str) -> Response:
        """The endpoint is used to update the tenant associated to the given id with details specified in the request body. To update custom properties or users, use the respective APIs.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
            name {str} -- Name of the tenant
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id
        headers = {'Content-Type': 'application/json'}
        payload = '{ "name" : "' + name + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def add_custom_property(self, tenant_id: str, key: str, value: str) -> Response:
        """The endpoint is used to add a custom property to the tenant associated to the given id.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
            key {str} -- Key of the custom property
            value {str} -- Value of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/customProperties'
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def delete_custom_property(self, tenant_id: str, custom_property_key: str) -> Response:
        """This endpoint is used to delete a custom property from the tenant associated to the given id.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
            custom_property_key {str} -- Key of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/customProperties/' + custom_property_key
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def update_custom_property(self, tenant_id: str, key: str, value: str) -> Response:
        """The endpoint is used to update a custom property of the tenant associated to the given id. The ‘key’ attribute cannot be modified.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
            key {str} -- Key of the custom property
            value {str} -- Value of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/customProperties/' + key
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def get_trusted_ca_certificates(self, tenant_id: str) -> Response:
        """The endpoint is used to download tenant specific trusted CA certificates for authentication.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/trustedCACertificates'
        return super().request_core(method='GET', service=service, accept_json=True)

    def get_gateway_registration_client_cert(self, tenant_id: str) -> Response:
        """The endpoint is used to list the fingerprints and expiration dates for gateway registration certificates of the given tenant.

        Arguments:
            tenant_id {str} -- Unique identifier of a tenant

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/gatewayRegistrations/clientCertificate'
        return super().request_core(method='GET', service=service, accept_json=True)

    def get_gateway_registration_client_cert_p12(self, tenant_id: str) -> Response:
        """The endpoint is used to download tenant specific p12 file for the registration of a gateway.

        Arguments:
            tenant_id {str} -- Unique identifier of a tenant

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/gatewayRegistrations/clientCertificate/p12'
        return super().request_core(method='GET', service=service, accept_json=True)

    def get_gateway_registration_client_cert_pem(self, tenant_id: str) -> Response:
        """The endpoint is used to download tenant specific pem file for the registration of a gateway.

        Arguments:
            tenant_id {str} -- Unique identifier of a tenant

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/gatewayRegistrations/clientCertificate/pem'
        return super().request_core(method='GET', service=service, accept_json=True)

    def revoke_gateway_cert(self, tenant_id: str, fingerprint: str) -> Response:
        """The endpoint is used to revoke a gateway registration certificate of the given tenant.

        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
            fingerprint {str} -- The fingerprint of the certificate hashed with SHA-256 in hex format.

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/gatewayRegistrations/clientCertificate/' + fingerprint
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def get_users(self, tenant_id: str, orderby=None, asc=True, skip=None, top=None) -> Response:
        """The endpoint returns a list of users assigned to the tenant associated to the given id.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
        
        Keyword Arguments:
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore not included in the result set. (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request. (default: {None})
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/users'
        query = build_query(orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=service, query=query, accept_json=True)

    def add_user(self, tenant_id: str, role: str, user_id: str) -> Response:
        """The endpoint is used to add the user specified in the request body to the tenant associated to the given id.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
            role {str} -- Role of the user
            user_id {str} -- ID of the user
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/users'
        headers = {'Content-Type': 'application/json'}
        payload = '{ "role" : "' + role + '", "userId" : "' + user_id + '" }'
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def delete_user(self, tenant_id: str, user_id: str) -> Response:
        """The endpoint is used to remove the user from the tenant associated to the given id.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
            user_id {str} -- Unique identifier of a user
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/users/' + user_id
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def get_user(self, tenant_id: str, user_id: str) -> Response:
        """The endpoint is used to return the tenant user associated to the given id.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
            user_id {str} -- Unique identifier of a user
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/users/' + user_id
        return super().request_core(method='GET', service=service, accept_json=True)

    def update_user(self, tenant_id: str, user_id: str, role: str) -> Response:
        """The endpoint is used to update the tenant user associated to the given id with details specified in the request body.
        
        Arguments:
            tenant_id {str} -- Unique identifier of a tenant
            user_id {str} -- Unique identifier of a user
            role {str} -- Role of the user
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + tenant_id + '/users/' + user_id
        headers = {'Content-Type': 'application/json'}
        payload = '{ "role" : "' + role + '", "userId" : "' + user_id + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)



""" Author: Philipp Steinrötter (steinroe) """

import ssl
import json
import uuid
import time
from datetime import datetime
import random
import base64
import paho.mqtt.client as mqtt

from .utils import current_milli_time


class PahoMQTT(mqtt.Client):
    """Overwrites the mqtt.Client class to work with password protected pem files
    """

    def tls_set(self, ca_certs=None, pemfile=None, secret=None, tls_version=None):
        """Configure network encryption and authentication options. Enables SSL/TLS support.
        ca_certs : a string path to the Certificate Authority certificate files
        that are to be treated as trusted by this client. If this is the only
        option given then the client will operate in a similar manner to a web
        browser. That is to say it will require the broker to have a
        certificate signed by the Certificate Authorities in ca_certs and will
        communicate using TLS v1, but will not attempt any form of
        authentication. This provides basic network encryption but may not be
        sufficient depending on how the broker is configured.
        By default, on Python 2.7.9+ or 3.4+, the default certification
        authority of the system is used. On older Python version this parameter
        is mandatory.
        pemfile is a string pointing to the PEM encoded client
        certificate. If the argument is not
        None it they will be used as client information for TLS based
        authentication.  Support for this feature is broker dependent. Note
        that if the file in encrypted and needs a password to
        decrypt it, you will have to provide the secret, too.
        tls_version allows the version of the SSL/TLS protocol used to be
        specified. By default TLS v1 is used. Previous versions (all versions
        beginning with SSL) are possible but not recommended due to possible
        security problems.
        Must be called before connect() or connect_async()."""
        if ssl is None:
            raise ValueError('This platform has no SSL/TLS.')

        if not hasattr(ssl, 'SSLContext'):
            # Require Python version that has SSL context support in standard library
            raise ValueError('Python 2.7.9 and 3.2 are the minimum supported versions for TLS.')

        # Create SSLContext object
        if tls_version is None:
            tls_version = ssl.PROTOCOL_TLSv1
            # If the python version supports it, use highest TLS version automatically
            if hasattr(ssl, "PROTOCOL_TLS"):
                tls_version = ssl.PROTOCOL_TLS
        context = ssl.SSLContext(tls_version)

        # Configure context
        if pemfile is not None:
            context.load_cert_chain(pemfile, password=secret)

        context.verify_mode = ssl.CERT_REQUIRED

        context.load_default_certs()

        self.tls_set_context(context)

        self.tls_insecure_set(False)


class MQTTClient(PahoMQTT):
    """Wrapper around the Paho MQTT Client to simplify its usage with the IoTS Cloud Gateway
    """

    def __init__(self, instance: str, device_alternate_id: str, pemfile: str, secret: str):
        """Instantiate MQTT Client configured for specified instance and device
        
        Arguments:
            instance {str} -- IoT Services instance
            device_alternate_id {str} -- The alternate id of the mqtt (router) device
            certfile_path {str} -- The certfile path for the mqtt (router) device
            keyfile_path {str} -- The keyfile path for the mqtt (router) device
        """
        super(MQTTClient, self).__init__(client_id=device_alternate_id)

        self.tls_set(tls_version=ssl.PROTOCOL_TLSv1_2, pemfile=pemfile, secret=secret)

        self.host = instance
        self.port = 8883
        self.device_alternate_id = device_alternate_id

        self._on_error = None
        self._on_command = None
        self._message_buffer = {}
        self._command_callbacks = {}

    @property
    def on_error(self):
        """If implemented, called when the client has error.
        Defined to allow error handling."""
        return self._on_error

    @on_error.setter
    def on_error(self, func):
        with self._callback_mutex:
            self._on_error = func

    @property
    def on_command(self):
        """If implemented, called when the client has received a command message.
        Defined to allow command handling."""
        return self._on_command

    @on_command.setter
    def on_command(self, func):
        with self._callback_mutex:
            self._on_command = func

    def connect(self, keepalive=60):
        """Connects to the broker
        
        Keyword Arguments:
            keepalive {int} -- The number of seconds the connection should be kept alive (default: {60})
        """
        super(MQTTClient, self).connect(host=self.host, port=self.port, keepalive=keepalive)
        self._subscribe_ack()

    def _subscribe_ack(self):
        service = 'ack/' + self.device_alternate_id
        self.message_callback_add(service, self._ack_message_handler)
        return super(MQTTClient, self).subscribe(service, 1)

    def _ack_message_handler(self, client, userdata, message):
        message_infos = json.loads(message.payload.decode("utf-8"))
        report = []
        for msg_info in message_infos:
            if msg_info.get('code') != 200 and msg_info.get('code') != 202:
                error = {
                    'message': self._message_buffer.get(msg_info.get('id')),
                    'error': ' '.join(msg_info.get('messages'))
                }
                report.append(error)
            del self._message_buffer[msg_info.get('id')]

        if len(report) > 0:
            self.on_error(self, userdata, report)

    def subscribe(self, device_alternate_id: str) -> (str, str):
        """Subscribe to a devices commands
        
        Arguments:
            device_alternate_id {str} -- The alternate id of the device
        
        Returns:
            str -- Result of the subscription
            str -- Message ID
        """
        service = 'commands/' + device_alternate_id
        self.message_callback_add(service, self._command_message_handler)
        return super(MQTTClient, self).subscribe(service, 1)

    def _command_message_handler(self, client, userdata, message):
        parsed_message = json.loads(message.payload.decode("utf-8"))
        self.on_command(self, userdata, parsed_message)

    def publish(self, capability_alternate_id: str, sensor_alternate_id: str, measures: list,
                device_alternate_id: str = None, timestamp: int = None) -> mqtt.MQTTMessageInfo:
        """Publishes measures to the IoT Services
        
        Arguments:
            device_alternate_id {str} -- Alternate ID of the device. If none, the device from the client id will be used.
            capability_alternate_id {str} -- Alternate ID of the capability
            sensor_alternate_id {str} -- Alternate ID of the sensor
            timestamp {int} -- UNIX time in milliseconds. If None, current time will be used.
            measures {list} -- List of key-value pairs containing the measures and their respective values 
        
        Returns:
            mqtt.MQTTMessageInfo -- MQTT Message Info
        """
        if device_alternate_id is None:
            device_alternate_id = self.device_alternate_id

        service = 'measures/' + device_alternate_id
        measure_message_id = str(uuid.uuid4())
        payload = {
            "timestamp": current_milli_time(),
            "capabilityAlternateId": capability_alternate_id,
            "sensorAlternateId": sensor_alternate_id,
            "measureMessageId": measure_message_id,
            "measures": measures
        }

        if timestamp is not None:
            payload['timestamp'] = timestamp

        self._message_buffer[measure_message_id] = payload
        payload_json = json.dumps(payload)
        return super(MQTTClient, self).publish(service, payload=payload_json)



""" Author: Philipp Steinrötter (steinroe) """

from .about import AboutService
from .capability import CapabilityService
from .device import DeviceService
from .gateway import GatewayService
from .protocol import ProtocolService
from .sensor_type import SensorTypeService
from .sensor import SensorService
from .tenant import TenantService
from .user import UserService
from .vendor import VendorService
from .session import SessionService

from .iot_service import IoTService, DeviceManagementAPIException
from .tenant_iot_service import TenantIoTService
from .rest_client import RestClient, RESTGatewayException
from .mqtt_client import MQTTClient

from .utils import debug_requests_off, debug_requests_on


""" Author: Philipp Steinrötter (steinroe) """

from .tenant_iot_service import TenantIoTService
from .utils import build_query
from .response import Response


class SensorService(TenantIoTService):
    def __init__(self,
                 instance,
                 user,
                 password,
                 tenant_id):
        """Instantiate SensorService object
        
        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
            tenant_id {string} -- Id of the tenant
        """

        self.service = '/sensors'

        TenantIoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password,
            tenant_id=tenant_id
        )

    def get_sensors(self, filters=None, orderby=None, asc=True, skip=None, top=None) -> Response:
        """The endpoint returns a list of sensors.
        
        Keyword Arguments:
            filters {list} -- This parameter allows clients to filter the collection for specific attributes. It is possible to filter by 'id’, 'deviceId’, 'name’, and 'alternateId’. The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"]. (default: {None})
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore included in the result set. (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request. (default: {None})
        
        Returns:
            Response -- Response object
        """
        query = build_query(filters=filters, orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=self.service, query=query, accept_json=True)

    def get_sensor_count(self):
        """The endpoint returns the count of all sensors.

        Returns:
            Response -- Response object
        """
        service = self.service + '/count'
        return super().request_core(method='GET', service=service, accept_json=True)

    def create_sensor(self, device_id: str, alternate_id: str, name: str, sensor_type_id: str) -> Response:
        """This endpoint is used to create a sensor.
        
        Arguments:
            device_id {str} -- Respective device ID for the sensor
            alternate_id {str} -- Alternate ID for the sensor
            name {str} -- Name for the sensor
            sensor_type_id {str} -- ID of the respective sensor type
        
        Returns:
            Response -- Response object
        """
        headers = {'Content-Type': 'application/json'}
        payload = '{ "deviceId": "' + device_id + '", "alternateId": "' + alternate_id + '", "name": "' + name + '", "sensorTypeId": "' + sensor_type_id + '"}'
        return super().request_core(method='POST', service=self.service, headers=headers, payload=payload,
                                    accept_json=True)

    def delete_sensor(self, sensor_id: str) -> Response:
        """The endpoint is used to delete the sensor associated to the given id.
        
        Arguments:
            sensor_id {str} -- Unique identifier of a sensor
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_id
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def get_sensor(self, sensor_id: str) -> Response:
        """The endpoint returns the sensor associated to the given id.
        
        Arguments:
            sensor_id {str} -- Unique identifier of a sensor
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_id
        return super().request_core(method='GET', service=service, accept_json=True)

    def update_sensor(self, sensor_id: str, name: str, sensor_type_id: str) -> Response:
        """This endpoint is used to update a sensor associated to the given id with details specified in the request body.
        
        Arguments:
            sensor_id {str} -- Unique identifier of a sensor
            name {str} -- Name of the sensor
            sensor_type_id {str} -- Respective sensor type ID
        
        Returns:
            Response -- [description]
        """
        service = self.service + '/' + sensor_id
        headers = {'Content-Type': 'application/json'}
        payload = '{ "name" : "' + name + '", "sensorTypeId" : "' + sensor_type_id + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def add_custom_property(self, sensor_id: str, key: str, value: str) -> Response:
        """The endpoint is used to add a custom property to the sensor associated to the given id.

        Arguments:
            sensor_id {str} --  Unique identifier of a sensor
            key {str} -- Key of the custom property
            value {str} -- Value of the custom property

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_id + '/customProperties'
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def delete_custom_property(self, sensor_id: str, key: str) -> Response:
        """Delete a custom property from the sensor associated to the given id.

        Arguments:
            sensor_id {str} -- Unique identifier of a sensor
            key {str} -- Key of the custom property

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_id + '/customProperties/' + key
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def update_custom_property(self, sensor_id: str, key: str, value: str) -> Response:
        """Updates a custom property of the sensor associated to the given id. The ‘key’ attribute cannot be modified.

        Arguments:
            sensor_id {str} -- Unique identifier of a sensor
            key {str} -- Key of the custom property
            value {str} -- The updated value of the custom property

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_id + '/customProperties/' + key
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)



""" Author: Philipp Steinrötter (steinroe) """

from .iot_service import IoTService
from .utils import build_query
from .response import Response


class VendorService(IoTService):
    def __init__(self,
                 instance,
                 user,
                 password):
        """Instantiate VendorService object
        
        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
        """

        self.service = '/vendors'

        IoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password
        )

    def get_vendors(self, skip=None, top=None) -> Response:
        """The endpoint returns a list of vendors, shared among all tenants of the instance.
        
        Keyword Arguments:
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore included in the result (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request (default: {None})
        
        Returns:
            Response -- Response object
        """
        query = build_query(skip=skip, top=top)
        return super().request_core(method='GET', service=self.service, query=query, accept_json=True)

    def get_vendor_count(self):
        """The endpoint returns the count of all vendors.

        Returns:
            Response -- Response object
        """
        service = self.service + '/count'
        return super().request_core(method='GET', service=service)

    def create_vendor(self, vendor_id: str) -> Response:
        """The endpoint is used to create a new vendor. The new vendor is visible for all tenants of the instance.
        
        Arguments:
            vendor_id {str} -- ID of the vendor that will be created
        
        Returns:
            Response -- Response object
        """
        headers = {'Content-Type': 'application/json'}
        payload = '{ "id": "' + vendor_id + '" }'
        return super().request_core(method='POST', service=self.service, headers=headers, payload=payload,
                                    accept_json=True)

    def delete_vendor(self, vendor_id: str) -> Response:
        """The endpoint is used to delete a vendor. Deleting is possible only if no tenant on the instance is using it anymore.
        
        Arguments:
            vendor_id {str} -- Unique identifier of a vendor
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + vendor_id
        return super().request_core(method='DELETE', service=service, accept_json=True)



""" Author: Philipp Steinrötter (steinroe) """

from .iot_service import IoTService
from .response import Response


class SessionService(IoTService):
    def __init__(self,
                 instance,
                 user,
                 password):
        """Instantiate SessionService object

        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
        """

        self.service = ''

        IoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password
        )

    def logout(self) -> Response:
        """Logs out the user by invalidating the session. The user is identified via session cookie or the
        Authorization header. """
        service = '/logout'
        response = self.request_core(method='POST', service=service, accept_json=True)
        return response

    def me(self) -> Response:
        """The current user is identified via session cookie or the Authorization header."""
        service = '/me'
        response = self.request_core(method='POST', service=service, accept_json=True)
        return response



from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name = 'iot_services_sdk',         
  packages = ['iot_services_sdk'],   
  version = '2.0.0',
  license='SAP Sample Code License',        
  description = 'SDK for SAP IoT Services on Cloud Foundry. Wraps all Device Management APIs as well as both Cloud Gateways.',  
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  python_requires='>=3',
  author = 'Philipp Steinrötter',                  
  author_email = 'philipp.steinroetter@sap.com',     
  url = 'https://github.com/SAP/iot-services-sdk',   
  keywords = 'SAP IoT Services CF SDK MQTT REST Device Management Internet of Things',
  download_url='https://github.com/SAP/iot-services-sdk/archive/1.0.tar.gz',
  install_requires=['paho-mqtt', 'requests'],
  classifiers=[
    'Development Status :: 5 - Production/Stable',      
    'Intended Audience :: Developers',   
    'Programming Language :: Python :: 3', 
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7'
  ],
)


""" Author: Philipp Steinrötter (steinroe) """

from .iot_service import IoTService
from .response import Response


class TenantIoTService(IoTService):
    def __init__(self, instance: str, user: str, password: str, tenant_id: str):
        """Instantiate IoT Service object

        Arguments:
            instance {str} -- IoT Service instance
            user {str} -- IoT Service user
            password {str} -- IoT Service password
            tenant {str} -- IoT Service Tenant

        Raises:
            ValueError -- Raised if any required argument is not provided
        """
        if instance is None or user is None or password is None or tenant_id is None:
            raise ValueError('You must specify your instance, user and password.')

        self._tenant_id = tenant_id

        self._service_base_path = '/tenant/' + tenant_id

        IoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password
        )

    @property
    def tenant_id(self):
        return self._tenant_id

    @tenant_id.setter
    def tenant_id(self, tenant_id):
        if tenant_id is None:
            raise ValueError('Tenant Id must be set. Please use IoTService base class to call a service not related '
                             'to a specific tenant.')

        self._tenant_id = tenant_id
        self._service_base_path = '/tenant/' + tenant_id

    def request_core(self, method=None, service=None, headers=None, payload=None, accept_json=False, query=None,
                     files=None) -> Response:
        """Fires a HTTP request to core services

        Keyword Arguments:
            method {str} -- HTTP method (default: {None})
            service {str} -- Service Path (default: {None})
            headers {dict} -- HTTP headers (default: {None})
            payload {str} -- Message payload (default: {None})
            accept_json {bool} -- If set to true, the response is parsed a JSON (default: {False})
            query {str} -- Query for filtering (default: {None})
            files {str} -- Path to files (default: {None})

        Returns:
            Response -- Response object
        """

        service = self._service_base_path + service
        return super().request_core(method=method, service=service, headers=headers, payload=payload, accept_json=accept_json, query=query, files=files)


""" Author: Philipp Steinrötter (steinroe) """

import json

from .tenant_iot_service import TenantIoTService
from .utils import build_query
from .response import Response


class SensorTypeService(TenantIoTService):
    def __init__(self,
                 instance,
                 user,
                 password,
                 tenant_id):
        """Instantiate SensorTypeService object
        
        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
            tenant_id {string} -- ID of the Tenant
        """
        self.service = '/sensorTypes'

        TenantIoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password,
            tenant_id=tenant_id
        )

    def get_sensor_types(self, filters=None, orderby=None, asc=True, skip=None, top=None) -> Response:
        """The endpoint returns a list of sensor types.
        
        Keyword Arguments:
            filters {list} -- This parameter allows clients to filter the collection for attributes of a sensorType. The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"] (default: {None})
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore not included in the result set. (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request. (default: {None})
        
        Returns:
            Response -- Response object
        """
        query = build_query(filters=filters, orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=self.service, query=query, accept_json=True)

    def get_sensor_type_count(self):
        """The endpoint returns the count of all sensor types.

        Returns:
            Response -- Response object
        """
        service = self.service + '/count'
        return super().request_core(method='GET', service=service, accept_json=True)

    def create_sensor_type(self, alternate_id: str, name: str, capabilities: list) -> Response:
        """This endpoint is used to create a sensor type.
        
        Arguments:
            alternate_id {str} -- Alternate ID of the sensor type
            name {str} -- Name of the sensor type
            capabilities {list} -- List of dicts each containing key-value pairs for 'id' and 'type'
        
        Returns:
            Response -- Response object
        """
        headers = {'Content-Type': 'application/json'}
        payload = json.dumps({"alternateId": alternate_id, "name": name, "capabilities": capabilities})
        return super().request_core(method='POST', service=self.service, headers=headers, payload=payload,
                                    accept_json=True)

    def delete_sensor_type(self, sensor_type_id: str) -> Response:
        """The endpoint is used to delete the sensor type associated to the given id.
        
        Arguments:
            sensor_type_id {str} -- Unique identifier of a sensor type
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_type_id
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def get_sensor_type(self, sensor_type_id: str) -> Response:
        """The endpoint returns the sensor type associated to the given id.
        
        Arguments:
            sensor_type_id {str} -- Unique identifier of a sensor type
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_type_id
        return super().request_core(method='GET', service=service, accept_json=True)

    def update_sensor_type(self, sensor_type_id: str, alternate_id: str, name: str) -> Response:
        """This endpoint is used to update the sensor type associated to the given id with details specified in the request body. To update capabilities, use the respective API.
        
        Arguments:
            sensor_type_id {str} -- Unique identifier of a sensor type
            alternate_id {str} -- Alternate identifier of the sensor type
            name {str} -- Name of the sensor type
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_type_id
        headers = {'Content-Type': 'application/json'}
        payload = '{ "alternateId" : "' + alternate_id + '", "name" : "' + name + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def add_capability(self, sensor_type_id: str, capability_id: str, capability_type: str) -> Response:
        """This endpoint is used to add a capability. Note that it is not supported to add a capability to a sensor type which is already associated with a sensor.
        
        Arguments:
            sensor_type_id {str} -- Unique identifier of a sensorType
            capability_id {str} -- ID of the capability that will be added
            capability_type {str} -- Type of the capability that will be added. Can be 'measure' or 'command'
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_type_id + '/capabilities'
        headers = {'Content-Type': 'application/json'}
        payload = '{ "id" : "' + capability_id + '", "type" : "' + capability_type + '" }'
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def remove_capability(self, sensor_type_id: str, capability_id: str) -> Response:
        """The endpoint is used to remove the capability associated to the given id. Note that it is not supported to delete a capability in a sensor type which is already associated with a sensor.
        
        Arguments:
            sensor_type_id {str} -- Unique identifier of a sensorType
            capability_id {str} -- Unique identifier of a capability
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_type_id + '/capabilities/' + capability_id
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def update_capability(self, sensor_type_id: str, capability_id: str, capability_type: str) -> Response:
        """This endpoint is used to update the capability associated to the given id with details specified in the request body. Note that it is not supported to modify the type of a capability in a sensor type if it is already associated with a sensor.
        
        Arguments:
            sensor_type_id {str} -- Unique identifier of a sensorType
            capability_id {str} -- ID of the capability that will be updated
            capability_type {str} -- Type of the capability that will be added. Can be 'measure' or 'command'
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + sensor_type_id + '/capabilities/' + capability_id
        headers = {'Content-Type': 'application/json'}
        payload = '{ "id" : "' + capability_id + '", "type" : "' + capability_type + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)



""" Author: Philipp Steinrötter (steinroe) """

import json

from .iot_service import IoTService
from .utils import build_query
from .response import Response


class UserService(IoTService):
    def __init__(self,
                 instance,
                 user,
                 password):
        """Instantiate UserService object
        
        Arguments:
            instance {string} -- IoT Services instance
            user {string} -- IoT Services user
            password {string} -- IoT Services password
        """

        self.service = '/users'

        IoTService.__init__(
            self,
            instance=instance,
            user=user,
            password=password
        )

    def get_users(self, filters=None, orderby=None, asc=True, skip=None, top=None) -> Response:
        """The endpoint returns a list of users.
        
        Keyword Arguments:
            filters {list} -- This parameter allows clients to filter the collection for attributes of a user.  The filters must be provided as a list of strings, e.q. ["name eq 'my-name'", "id eq '111'"]. (default: {None})
            orderby {str} -- The attribute to order by. (default: {None})
            asc {bool} -- Only considered if orderby is not none. Defines if the values should be ordered asc or desc. (default: {True})
            skip {int} -- This parameter specifies the number of items in the queried collection which will be skipped and therefore not included in the result (default: {None})
            top {int} -- This parameter restricts the maximum number of items which will be returned by the request (default: {None})
        
        Returns:
            Response -- Response object
        """
        query = build_query(filters=filters, orderby=orderby, asc=asc, skip=skip, top=top)
        return super().request_core(method='GET', service=self.service, query=query, accept_json=True)

    def create_user(self, name: str, password: str, custom_properties=[]) -> Response:
        """The endpoint is used to create a user. Note: This function only supports basic authentication method.
        
        Arguments:
            name {str} -- Name of the user
            password {str} -- Password of the user
            custom_properties {list} -- Custom properties for the user given as list of dicts each with the key-value pairs 'key' and 'value'
        
        Returns:
            Response -- Response object
        """
        headers = {'Content-Type': 'application/json'}
        payload = json.dumps({"name": name, "customProperties": custom_properties,
                              "authentications": [{"type": "basic", "password": password}]})
        return super().request_core(method='POST', service=self.service, headers=headers, payload=payload,
                                    accept_json=True)

    def get_user_count(self) -> Response:
        """The endpoint is used to delete the user associated to the given id.

        Returns:
            Response -- Response object
        """
        service = self.service + '/count'
        return super().request_core(method='GET', service=service, accept_json=True)

    def delete_user(self, user_id: str) -> Response:
        """The endpoint is used to delete the user associated to the given id.
        
        Arguments:
            user_id {str} -- Unique identifier of a user
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + user_id
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def get_user(self, user_id: str) -> Response:
        """The endpoint returns the user associated to the given id.
        
        Arguments:
            user_id {str} -- Unique identifier of a user
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + user_id
        return super().request_core(method='GET', service=service, accept_json=True)

    def update_password(self, user_id: str, password: str) -> Response:
        """The endpoint is used to update the password of the user associated to the given id.
        
        Arguments:
            user_id {str} -- Unique identifier of a user
            password {str} -- Password that will be updated
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + user_id + '/authentications/basic'
        headers = {'Content-Type': 'application/json'}
        payload = '{ "password": "' + password + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def add_custom_property(self, user_id: str, key: str, value: str) -> Response:
        """The endpoint is used to add a custom property to the user associated to the given id.
        
        Arguments:
            user_id {str} -- Unique identifier of a user
            key {str} -- Key of the custom property that will be created
            value {str} -- Value of the custom property that will be created
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + user_id + '/customProperties'
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def delete_custom_property(self, user_id: str, key: str) -> Response:
        """The endpoint is used to delete a custom property from the user associated to the given id.
        
        Arguments:
            user_id {str} -- Unique identifier of a user
            key {str} -- Key of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + user_id + '/customProperties/' + key
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def update_custom_property(self, user_id: str, key: str, value: str) -> Response:
        """The endpoint is used to update a custom property of the user associated to the given id. The ‘key’ attribute cannot be modified.
        
        Arguments:
            user_id {str} -- Unique identifier of a user
            key {str} -- Key of the custom property
            value {str} -- Value of the custom property
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + user_id + '/customProperties/' + key
        headers = {'Content-Type': 'application/json'}
        payload = '{ "key" : "' + key + '", "value" : "' + value + '" }'
        return super().request_core(method='PUT', service=service, headers=headers, payload=payload, accept_json=True)

    def add_role(self, user_id: str, role: str) -> Response:
        """The endpoint is used to add a role to user associated to the given id. The role is valid across the instance.
        
        Arguments:
            user_id {str} -- Unique identifier of a user
            role {str} -- Specification of the role that will be created
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + user_id + '/roles'
        headers = {'Content-Type': 'application/json'}
        payload = '{ "role" : "' + role + '" }'
        return super().request_core(method='POST', service=service, headers=headers, payload=payload, accept_json=True)

    def delete_role(self, user_id: str, role: str) -> Response:
        """The endpoint is used to delete the role from user associated to the given id. The role is valid across the instance.
        
        Arguments:
            user_id {str} -- Unique identifier of a user
            role {str} -- Name of the role
        
        Returns:
            Response -- Response object
        """
        service = self.service + '/' + user_id + '/roles/' + role
        return super().request_core(method='DELETE', service=service, accept_json=True)

    def get_tenants(self, user_id: str) -> Response:
        """The endpoint returns a list of tenants assigned to the user associated with the given id.

        Arguments:
            user_id {str} -- Unique identifier of a user

        Returns:
            Response -- Response object
        """
        service = self.service + '/' + user_id + '/tenants'
        return super().request_core(method='GET', service=service, accept_json=True)
